"""
Definitions and data useful for many services modules.

"""
# warning: do NOT remove any of these definitions, even though they are not used in this module.
from google.auth.credentials import Credentials as GcpCredentials  # noqa: F401
from googleapiclient.discovery import Resource as GcpResource  # noqa: F401
